"""
Define some wrappers to help generate a collection of different populations
that might be useful to compare with one another. This is mostly a tool

"""

from ..imports import *
from .exoplanets import *
from .solarsystem import *


def get_transiting_and_nontransiting_exoplanets():
    e = Exoplanets()
    transit = e[e.detected_in_transit() != 0]
    transit.label = "Transiting Exoplanets"
    nontransit = e[e.detected_in_transit() == 0]
    nontransit.label = "Non-transiting Exoplanets"
    return dict(nontransit=nontransit, transit=transit)


def get_exoplanets_by_method(methods="all", include_solar_system=True):
    """
    Create a dictionary that contains a collection of exoplanet
    populations, separated by their discovery method.

    Returns
    -------

    pops : dict
        A dictionary of planet populations,
        grouped by discovery method.
    """

    # create a population of exoplanets
    e = Exoplanets()

    # use all the methods
    if methods == "all":
        methods = np.unique(e.discovery_method())

    # create a dictionary of populations,
    pops = {}
    for m in methods:
        pops[m] = e[e.discovery_method() == m]
        pops[m].label = m
        pops[m].color = None

    if include_solar_system:
        s = SolarSystem()
        s.color = "black"
        pops["Solar System"] = s

    return pops


def get_solar_system_objects():
    """
    Create a dictionary that contains a collection of
    Solar System populations, grouped by category.

    Returns
    -------
    pops : dict
        A dictionary of Solar System populations,
        grouped by (JPL/SSD) category.
    """
    pops = dict(
        major=SolarSystem(),
        dwarf=SolarSystemDwarfPlanets(),
        moons=SolarSystemMoons(),
        minor=SolarSystemMinorPlanets(),
    )
    return pops


def get_transiting_exoplanets_by_mass_precision(sigma=5):
    """
    Create a dictionary that contains a collection of exoplanet
    populations, separated by whether their masses are good or bad.

    Parameters
    ----------
    sigma : float
        At how many sigma should the mass be
        distinguishable from 0? The maximum
        fractional uncertainty on planets
        classified as 'good' will be 1/sigma.
    Returns
    -------
    pops : dict
        A dictionary of planet populations,
        grouped by mass precision.
    """
    # create a dictionary of populations,
    pops = dict(good=GoodMass(sigma=sigma), bad=BadMass(sigma=sigma))
    return pops


def get_exoplanets_by_teff():
    """
    Create a dictionary that contains a collection of exoplanet
    populations, separated by their stellar effective temperature.

    Returns
    -------
    pops : dict
        A dictionary of planet populations,
        grouped by stellar effective temperature.
    """

    e = Exoplanets()
    # spectral type boundaries from https://www.pas.rochester.edu/~emamajek/EEM_dwarf_UBVIJHK_colors_Teff.txt
    upper_temperature_limits = dict(
        Y=480,
        T=1312,
        L=2310,
        M=3890,
        K=5325,
        G=5960,
        F=7310,
        A=10050,
        B=31650,
        O=np.inf,
    )

    pops = {}
    for k, t in upper_temperature_limits.items():
        is_cool_enough = e.stellar_teff() < t * u.K
        try:
            pops[k] = e[is_cool_enough]
            pops[k].label = k
            pops[k].color = None
        except IndexError:
            pass
        e = e[is_cool_enough == False]

    # reverse order to get OBAFGKMLTY
    items = list(pops.items())
    items.reverse()
    return dict(items)


def get_all_planets():
    p = get_transiting_and_nontransiting_exoplanets() | get_solar_system_objects()
    p["transit"].color = "black"
    p["nontransit"].color = "darkorchid"
    p["major"].annotate_planets = True
    p["dwarf"].annotate_planets = True
    return p



from .Map import *

__all__ = ["BubbleMap"]

default_size = plt.matplotlib.rcParams["lines.markersize"] ** 2


class BubbleMap(Map):
    """
    BubbleMap is a general wrapper for making scatter plots
    where planets are represented as bubbles that can have
    informative sizes and/or colors. As bubbles can have
    x, y, size, color; a BubbleMap can be a decent way
    to represent up to four different dimensions.
    """

    color = None
    size = None

    def __init__(
        self,
        xaxis=None,
        yaxis=None,
        size=None,
        color=None,
        size_normalization=10,
        size_vmin=None,
        size_vmax=None,
        size_scale=None,
        cmap="plasma",
        color_vmin=None,
        color_vmax=None,
        color_scale=None,
        **kw,
    ):
        """
        Initialize a plotting map for drawing bubbles.

        Parameters
        ----------
        xaxis : Plottable, str, None
            What should the x-positions encode?
        yaxis : Plottable, str, None
            What should the y-positions encode?
        size : Plottable, str, float, None
            What should the sizes of points be or encode?
        color : Plottable, str, float, None
            What should the colors of points be or encode?
        size_normalization : float
            If sizes depend on quantities, how big should
            the symbols be for a `normalized_value` of 1?
            Sizes are scaled relative to the default
            symbol area in `matplotlib`, so a value of
            `size_normalization=1` will give default areas.
        size_vmin : float, None
            If the sizes depend on quantities, what
            unnormalized data value corresponds to the
            a symbol area of 0? (This will overwrite
            `lim` set for the size plottable.)
        size_vmax : float, None
            If the sizes depend on quantities, what
            unnormalized data value corresponds to the
            default symbol area? (This will overwrite
            `lim` set for the size plottable.)
        size_scale: float, None
            If the sizes depend on quantities, what
            scaling should be applied when normalizing
            values? Options are `linear` and `log`.
            (This will overwrite `scale` set for the
            size plottable.)
        cmap : str, cmap from plt.matplotlib.cm
            If the colors depend on quantities, what cmap
            should be used for them? Colors will be based
            on `normalized_value` from the Plottable
        color_vmin : float, None
            If the colors depend on quantities, what
            unnormalized data value corresponds to the
            bottom of the color map? (This will overwrite
            `lim` set for the color plottable.)
        color_vmax : float, None
            If the colors depend on quantities, what
            unnormalized data value corresponds to the
            top of the color map? (This will overwrite
            `lim` set for the color plottable.)
        color_scale: float, None
            If the colors depend on quantities, what
            scaling should be applied when normalizing
            values? Options are `linear` and `log`.
            (This will overwrite `scale` set for the
            color plottable.)
        **kw : dict
            Other keywords will be passed on to *all*
            Map/Plottable initializations (which may
            include x, y, size, and color). If you need
            more fine-grained control over which axis
            gets which keyword, consider initializing
            those maps one-by-one.
        """

        # initialize the basics of the map with the plottable axes
        Map.__init__(self, xaxis=xaxis, yaxis=yaxis, **kw)

        # set up how we should scale the sizes of points
        size_to_use = clean_plottable(size or self.size, **kw)
        if isinstance(size_to_use, Plottable):
            self.plottable["size"] = size_to_use
            # overwrite plottable defaults, if requested
            if size_vmin is not None:
                self.plottable["size"].lim[0] = size_vmin
            if size_vmax is not None:
                self.plottable["size"].lim[1] = size_vmax
            if size_scale is not None:
                self.plottable["size"].scale = size_scale
        else:
            self.static_size = size_to_use

        # make sure a size normalization has been defined (relative to default)
        self.size_normalization = size_normalization

        # set up how we should set the colors of points
        color_to_use = clean_plottable(color or self.color, **kw)
        if isinstance(color_to_use, Plottable):
            self.plottable["color"] = color_to_use
            # overwrite plottable defaults, if requested
            if color_vmin is not None:
                self.plottable["color"].lim[0] = color_vmin
            if color_vmax is not None:
                self.plottable["color"].lim[1] = color_vmax
            if color_scale is not None:
                self.plottable["color"].scale = color_scale
        else:
            self.static_color = color_to_use

        # if an actual cmap was provided, use it
        if isinstance(cmap, plt.matplotlib.colors.Colormap):
            self.cmap = cmap
        # otherwise, treat the cmap as a string key
        else:
            self.cmap = plt.matplotlib.colormaps[cmap]

        # apply (x,y) axis labels, scales, limits appropriately
        for axis in "xy":
            for attribute in ["label", "scale", "lim"]:
                setattr(
                    self, f"{axis}{attribute}", getattr(self.plottable[axis], attribute)
                )

    def get_sizes(self):
        """
        The sizes of the bubbles.

        Returns
        -------
        s : an input for plt.scatter
            Either a single scalar, or an array with variable
            sizes for each bubble according to some quantity.
        """

        # should we ignore any variable size instructions?
        population_size = self.pop._plotkw.get("size", None) or self.pop._plotkw.get(
            "s", None
        )
        if self.pop.respond_to_size == False:
            size = population_size or self.static_size
        # if desired, set variable sizes
        elif "size" in self.plottable:
            # get the raw values for the sizes
            x = self.plottable["size"].normalized_value(self.pop)
            # calculate the normalized size
            size = default_size * self.size_normalization * x
        # otherwise, set a single size
        else:
            # get default, first from pop and then from map
            size = population_size or self.static_size
        # return a valid input to plt.scatter(s=...)
        return size

    def get_colors(self):
        """
        The colors of the bubbles.

        Returns
        -------
        c : an input for plt.scatter
            Either a single color, or an array with variable
            colors for each bubble according to some quantity.
        """

        # should we ignore any variable color instructions?
        population_color = self.pop._plotkw.get("color", None) or self.pop._plotkw.get(
            "c", None
        )
        if self.pop.respond_to_color == False:
            color = population_color or self.static_color
        # should we use a variable color?
        elif "color" in self.plottable:
            normalized = self.plottable["color"].normalized_value(self.pop)
            color = self.cmap(normalized)
        # finally, should we just use a default color?
        else:
            # get default, first from pop and then from map
            color = population_color or self.static_color

        # return a valid input to any one of the following:
        #   plt.scatter(c=...)
        #   plt.scatter(edgecolors=...)
        #   plt.scatter(facecolors=...)
        return color

    def kw(self, pop=None, **kw):
        """
        Do a little decision-making about the plotting keyword
        arguments, pulling defaults from each population where
        needed.

        Parameter
        ---------
        pop : Population, str, int
            The population to plot. This could be either an
            actual Population object, or a key referring to
            an element of the `self.populations` dictionary.
        **kw : dict
            All other keywords will be directed toward
            overwriting individual population defaults.
        """

        # make sure we're pointing at the right population
        if pop != None:
            self.point_at(pop)

        # define some default keywords, which can be over-written
        default = dict(
            s=self.get_sizes(),
            marker=self.pop.marker,
            linewidth=self.pop.linewidth,
            alpha=self.pop.alpha,
            zorder=self.pop.zorder,
            label=self.pop.label,
        )

        # sort out whether faces and/or edges should get color
        c = self.get_colors()
        if self.pop.filled:
            default["facecolors"] = c
        else:
            default["facecolors"] = "none"
        if self.pop.outlined:
            default["edgecolors"] = c
        else:
            default["edgecolors"] = "none"

        # if any other keywords are provided, overwrite these defaults
        default.update(**kw)

        return default

    '''def plot(self, key, ax=None, annotate_kw={}, **kw):
        """
        Add the points for a particular population to this map.

        Parameters
        ----------
        key : str
            The population (as an item in the self.populations dictionary) to add.
        ax :
            Into what ax should we place this plot?
            If None, use default.
        annotate_kw : dict
            Keywords for labeling the planet names.
        **kw : dict
            Any extra keywords will be passed on to `scatter`
        """

        # focus attention on that population
        self.point_at(key)

        # make sure we're plotting into the appropriate axes
        try:
            plt.sca(self.ax)
        except AttributeError:
            self.setup_axes(ax=ax)

        # add the scattered points
        self.scattered[key] = self.ax.scatter(self.x, self.y, **self.kw(key, **kw))

        # set the scales, limits, labels
        self.finish_plot(annotate_kw=annotate_kw)
'''



from .setup_tests import *
from exoatlas.imports import *
from exoatlas import *
from exoatlas.telescopes import *
from exoatlas.visualizations import *


def test_telescope_units():
    for t in ["Kepler", "TESS", "JWST", "HST"]:
        define_telescope_unit_by_name(t)
        define_telescope_unit_by_name(t, wavelength=0.7 * u.micron)


def test_buckets():
    """
    Make sure that our photon-counting tools plottables work,
    with their various possible telescope units.
    """

    t = TransitingExoplanets()[:20]
    Depth_x_Brightness().build([t])
    BubbleMap(xaxis=StellarBrightness, yaxis=Depth).build([t])
    BubbleMap(xaxis=StellarBrightness(wavelength=5 * u.micron), yaxis=Depth).build([t])
    for k in telescope_units:
        BubbleMap(
            xaxis=StellarBrightnessTelescope(telescope_name=k), yaxis=Depth
        ).build([t])
    BubbleMap(
        xaxis=StellarBrightnessTelescope(
            telescope_name="JWST", wavelength=10 * u.micron, R=10
        ),
        yaxis=Depth,
    ).build([t])


if __name__ == "__main__":  # pragma: no cover
    outputs = {k.split("_")[-1]: v() for k, v in locals().items() if "test_" in k}



from ..imports import *

from scipy.stats.mstats import mquantiles
from scipy.optimize import fsolve
import scipy.stats as STS

from scipy.signal import medfilt
from astropy.table import Table


def confidenceInterval(values, interval=0.682689, ThSig=False):
    "For determining the median and central confidence interval of input values array"
    if ThSig:
        interval = 0.997300203937  # 3 sigma
    qL = (1.0 - interval) / 2.0
    qU = 1.0 - qL
    out = mquantiles(values, prob=[qL, 0.5, qU])
    return out


from scipy.stats import skewnorm, norm

gaussian_central_1sigma = np.diff(norm(loc=0, scale=1).cdf([-1, 1]))[0]


def plot_skewnormal(mu=0, sigma=1, alpha=0):
    fi, ax = plt.subplots(2, 1, sharex=True, constrained_layout=True, figsize=(7, 3))
    p = skewnorm(a=alpha, loc=mu, scale=sigma)
    x = np.linspace(mu - 5 * sigma, mu + 5 * sigma, 1000)
    skew_median = p.median()
    skew_lower, skew_upper = p.interval(gaussian_central_1sigma)
    skew_mode = x[np.argmax(p.pdf(x))]

    median_description = rf"""
    median = {skew_median:.2f}  
    median - {gaussian_central_1sigma/2:.2%} = {skew_median - skew_lower:.2f}  
    {gaussian_central_1sigma/2:.2%} - median = {skew_upper - skew_median:.2f}  
    upper/lower = {(skew_upper - skew_median)/(skew_median - skew_lower):.2f}  
    """

    mode_description = rf"""
    mode = {skew_mode:.2f}  
    mode - {gaussian_central_1sigma/2:.2%} = {skew_mode - skew_lower:.2f}  
    {gaussian_central_1sigma/2:.2%} - mode = {skew_upper - skew_mode:.2f}  
    upper/lower = {(skew_upper - skew_mode)/(skew_mode - skew_lower):.2f}  
    """

    plt.sca(ax[0])
    plt.plot(x, p.pdf(x))
    plt.text(
        0,
        1,
        median_description,
        ha="left",
        va="top",
        transform=ax[0].transAxes,
        color="royalblue",
    )
    plt.text(
        1,
        1,
        mode_description,
        ha="right",
        va="top",
        transform=ax[0].transAxes,
        color="purple",
    )

    plt.ylabel("p(x)")
    plt.title(f"\nmu = {mu:.2f}, sigma = {sigma:.2f}, alpha = {alpha:.2f}")
    plt.sca(ax[1])
    plt.plot(x, p.cdf(x))
    plt.ylabel(r"$\int_0^{x} p(x') dx$")

    for a in ax:
        plt.sca(a)
        plt.axvline(skew_median, color="royalblue")
        plt.axvline(skew_mode, color="purple")

        plt.axvspan(skew_lower, skew_upper, alpha=0.3, color="gray")


def estSkewNorm(xin, conf=(0.16, 0.5, 0.84), Guess=None, Mode="Med", Check=True):
    """
    Used for estimating the parameters of a Skew Normal distribution that matches input mode/median and given confidence interval.

    Numerically solves system of equations to give the 3 output parameters of corresponding skew normal. May require multiple guess to result in right solution. Output should give practically exact values, see `Check'.

    Returns the skew normal parameters (mu, sigma, alpha) --- see wikipedia on skew normal

    WARNING: Code is provided with minimal explanation, and verfication --- use with that in mind.

    ZKBT borrowed this amazing code from Sebastian Pineda's
    https://github.com/jspineda/stellarprop/
    on 13 November 2024. If you want to understand it more and/or use it, please see Appendix B
    of https://ui.adsabs.harvard.edu/abs/2021ApJ...918...40P/abstract, and cite that paper!

    >---

    Parameters
    ----------
    xin : array_like, tuple, or list, with 3 entries
        entries correspond to the data value matching the corresponding index in the `conf' key word. By default first entry indicates the lower bound for the central 68% confidence interval, second entry corresponds to median/mode value depending on keyword `Mode', and the third entry is the upper bound for the central 68% confidence interval.

    conf : array_like, tuple, or list, with 3 entries
        indicates the values that the skew normal cumulative probability distribution should give with input `xin'. By default, set to median and central 68% confidence interval. If Mode is `Peak' the median equation is replaced by one corresponding to peak of distribution.

    Guess : array_like, tuple or list, with 3 entries
        allows for user to input starting point for numerical equation solve. Default values are a general guess. If output does not converge, use Guess to change starting point for computation. May require iteration for adequete solution. Use `Check' to verify. If there is difficult, input parameters may not be well suited for approximation with skew normal.

    Mode : str one of ['Peak','Med2','Med','SF']
        Defines to set of equations used in defining the skew normal distribution. If 'Peak' system sets second entry to be the mode of skew normal instead of median. All others are for setting the median, but with slightly different numerical implementations. 'Peak' and 'Med' are the recommended modes.

    Check : boolean
        By default it is True. Used as verification on output solution, gives printed diagnostics as check. Outputs for converged solutions should be exact if fully successful.


    Returns
    -------

    out : array_like with 3 entries
        gives the (mu, sigma, alpha) parameters that define the skew normal distribution matching inputs

    Notes
    -----
    Printed warnings also given from scipy from fsolve to diagnose progress of numerical solution to system of equations

    Examples
    --------

    ## Note that here we use data from https://github.com/jspineda/stellarprop for illustration ; see also Pineda et al. 2021b

    >> trace = np.genfromtxt('../resources/massradius/fractional_01/chains.csv',delimiter=',',names=True)
    >> scatlb, scatmid, scatub = confidenceInterval(trace['scatter'])  # the scatter psoterior distribution is asymetric, these are typically the reported values in literature
    >> print([scatlb,scatmid,scatub])
    [0.02787424918238516, 0.0320051813038165, 0.03692976181631807]
    >> params = estSkewNorm( [scatlb, scatmid, scatub])
    Mode at [0.03121118]
    Median at 0.032005181304171265
    Result gives centeral 68.0% confidence interval: (0.027874249182851436, 0.03692976181636316)
    Peak at [0.03121118] - [0.00333693]  +  [0.00571858]
    >> print(params)
    [0.02771848 0.0065575  1.95731243]

    ## Note that Check outputs reported numerically match nearly exactly to inputs, these would be kinda off if iteration needed
    ## In this example alpha ~2, indicating positive skewness, peak (mode) is at 0.031, a little less than median at 0.032   -- see appendix of Pineda et al. 2021b


    """

    xl, x0, xu = xin
    cl, c0, cu = conf
    if Guess is not None:
        p0 = Guess
    else:
        p0 = (x0, (xu - xl) / 2.0, ((xu - x0) - (x0 - xl)) / ((xu - xl) / 2.0) * 3)

    ## if block used to toggle set of equations to solve using scipy fsolve
    if Mode == "Peak":
        # print("Setting Peak of Distribution")
        # KLUDGE!!!!!
        mu_guess = x0
        sigma_guess = np.sqrt((xu - x0) ** 2 + (x0 - xl) ** 2)
        alpha_guess = np.log((xu - x0) / (x0 - xl)) * 3
        p0 = mu_guess, sigma_guess, alpha_guess

        def eq_sys(p):
            mu, sigma, alpha = p
            t = (x0 - mu) / sigma
            return (
                STS.skewnorm.cdf(xl, alpha, mu, sigma) - conf[0],
                STS.skewnorm.cdf(xu, alpha, mu, sigma) - conf[2],
                alpha * STS.norm.pdf(alpha * t) - STS.norm.cdf(alpha * t) * t,
            )

    elif Mode == "Med2":
        # print("Setting Median of Distribution")

        def eq_sys(p):
            mu, sigma, alpha = p
            return np.power(STS.skewnorm.cdf(xin, alpha, mu, sigma) - np.array(conf), 2)

    elif Mode == "SF":
        # print("Setting Median of Distribution")

        def eq_sys(p):
            mu, sigma, alpha = p
            return STS.skewnorm.isf(1 - np.array(conf), alpha, mu, sigma) - np.array(
                xin
            )

    elif Mode == "Med":
        # print("Setting Median of Distribution")

        def eq_sys(p):
            mu, sigma, alpha = p
            return (
                STS.skewnorm.cdf(xl, alpha, mu, sigma) - cl,
                STS.skewnorm.cdf(x0, alpha, mu, sigma) - c0,
                STS.skewnorm.cdf(xu, alpha, mu, sigma) - cu,
            )

    out, infodict, ier, message = fsolve(eq_sys, p0, factor=0.1, full_output=True)
    ok = bool(ier)
    mu, sigma, alpha = out

    if Check:
        ff = lambda a: STS.norm.pdf(out[2] * a) * out[2] - a * STS.norm.cdf(a * out[2])
        tm = fsolve(ff, 0.2 * out[2])
        xm = tm * out[1] + out[0]
        print("Mode at {}".format(xm))
        print("Median at {}".format(STS.skewnorm.median(out[2], out[0], out[1])))
        print(
            "Result gives centeral {0}% confidence interval:".format(
                (conf[2] - conf[0]) * 100
            ),
            STS.skewnorm.interval(conf[2] - conf[0], out[2], out[0], out[1]),
        )
        print("Peak at {0} - {1}  +  {2} ".format(xm, xm - xl, xu - xm))

        # ZKBT adds a plot
        plot_skewnormal(mu, sigma, alpha)

    return mu, sigma, alpha, ok


def make_skewnormal_parameters_to_interpolate(N=100):
    """
    Make a table to convert from sigma_upper/sigma_lower to skewnormal parameters.

    This function is/was used to generate the table in
    exoatlas/populations/data/skewnormal_parameters_for_interpolating_mode.ecsv
    that gets loaded whenever exoatlas is imported, to be
    able to quickly create skewnormal distributions for
    quantities with asymmetric uncertainties.

    """
    center = 0
    sigma_lower = 1
    sigmas_upper = np.logspace(-2, 2, N)
    tables = {}
    mylabels = {"Med": "median", "Peak": "mode"}
    for mode in ["Peak", "Med"]:
        iterations = 2

        x0 = center
        xu = center + sigmas_upper
        xl = center - sigma_lower
        mu_guess = x0 * np.ones_like(sigmas_upper)
        sigma_guess = np.sqrt((xu - x0) ** 2 + (x0 - xl) ** 2)
        if mode == "Peak":
            alpha_guess = np.log((xu - x0) / (x0 - xl)) * 3  # 100
        elif mode == "Med":
            alpha_guess = ((xu - x0) - (x0 - xl)) / (xu - xl) * 100
        guesses = zip(mu_guess, sigma_guess, alpha_guess)

        for i in range(iterations):
            parameters = np.array(
                [
                    estSkewNorm(
                        [center - sigma_lower, center, center + sigma_upper],
                        Mode=mode,
                        Guess=guess,
                        Check=False,
                    )
                    for sigma_upper, guess in zip(sigmas_upper, guesses)
                ]
            ).T
            ok = parameters[3].astype(bool)
            parameters = parameters[:, ok]

            mu_skew = parameters[0]
            sigma_skew = parameters[1]
            alpha_skew = parameters[2]

            filter_size = 11

            plt.figure()
            plt.plot(
                sigmas_upper,
                mu_skew,
                label=r"skew $\mu$",
                color="red",
                alpha=0.3,
                linewidth=3,
            )
            plt.plot(
                sigmas_upper,
                sigma_skew,
                label=r"skew $\sigma$",
                color="green",
                alpha=0.3,
                linewidth=3,
            )
            plt.plot(
                sigmas_upper,
                alpha_skew,
                label=r"skew $\alpha$",
                color="blue",
                alpha=0.3,
                linewidth=3,
            )
            plt.xscale("log")
            plt.legend(frameon=False)
            plt.xlabel(r"$\sigma_{upper}$")
            plt.title(
                f"mode = {mode}; "
                + r"$\mu_{original} = $"
                + f"{center}; "
                + r"$\sigma_{lower} = $"
                + f"{sigma_lower} | "
                + f"iteration #{i}"
            )

            plt.plot(sigmas_upper, mu_guess, color="red", linestyle="--")
            plt.plot(sigmas_upper, sigma_guess, color="green", linestyle="--")
            plt.plot(sigmas_upper, alpha_guess, color="blue", linestyle="--")
            plt.ylim(-30, 30)

            # update guess
            mu_guess = medfilt(mu_skew, filter_size)
            sigma_guess = medfilt(sigma_skew, filter_size)
            alpha_guess = medfilt(alpha_skew, filter_size)
            guesses = zip(mu_guess, sigma_guess, alpha_guess)

        t = Table(
            dict(
                sigma_upper=sigmas_upper,
                mu_skew=mu_guess,
                sigma_skew=sigma_guess,
                alpha_skew=alpha_guess,
            )
        )
        t.meta["mu"] = 0
        t.meta["sigma_lower"] = 1
        filename = f"skewnormal_parameters_for_interpolating_{mylabels[mode]}.ecsv"
        t.write(
            filename,
            overwrite=True,
        )
        print(f"saved parameter interpolating table to {filename}")
        tables[mylabels[mode]] = t
    return tables


interpolation_filename = os.path.join(
    code_directory, "populations/data/skewnormal_parameters_for_interpolating_mode.ecsv"
)
skew_table = ascii.read(interpolation_filename)


def make_skew_samples_from_lowerupper(
    mu=0, sigma_lower=1, sigma_upper=1, N_samples=100
):
    """
    Generate a skew-normal distribution from a mode and asymmetric error bars.

    Parameters
    ----------
    mu : float, Quantity, and/or array-like
        The mode of the distribution(s) = the peak of the distribution.
        For nearly symmetric distributions the mode and mean and median are
        all very close, but for more skewed distributions this is making a
        (slightly arbitrary) strong choice about distribution shapes.

    sigma_lower : float, Quantity, and/or array-like
        The magnitude of the lower uncertainty = the distance from the mode
        mu down to the 15.87% confidence interval, which would be exactly
        1 sigma below the mode/median/mean of symmetric Gaussian.

    sigma_upper: float, Quantity, and/or array-like
        The magnitude of the upper uncertainty = the distance from the mode
        mu down to the 15.87% confidence interval, which would be exactly
        1 sigma below the mode/median/mean of symmetric Gaussian.

    N_samples : int
        The number of samples to generate.

    Returns
    -------


    """

    assert np.max([len(np.shape(x)) for x in [mu, sigma_lower, sigma_upper]]) <= 1

    N_data = np.max([np.size(mu), np.size(sigma_lower), np.size(sigma_upper)])

    # find coefficients for mu=0, sigma_lower=1, sigma_upper/sigma_lower
    mu_skew = np.interp(
        sigma_upper / sigma_lower, skew_table["sigma_upper"], skew_table["mu_skew"]
    )
    sigma_skew = np.interp(
        sigma_upper / sigma_lower, skew_table["sigma_upper"], skew_table["sigma_skew"]
    )
    alpha_skew = np.interp(
        sigma_upper / sigma_lower, skew_table["sigma_upper"], skew_table["alpha_skew"]
    )

    # populate empty array with nans, to skip asking skewnormal to handle them
    ok = np.isfinite(mu_skew * sigma_skew * alpha_skew)
    samples = np.ones((N_samples, N_data)) * mu[np.newaxis, :] * np.nan

    # create zero-centered, unnormalized skewnormal distribution + samples
    p = skewnorm(loc=mu_skew[ok], scale=sigma_skew[ok], a=alpha_skew[ok])
    samples[:, ok] = p.rvs(size=(N_samples, sum(ok))) * sigma_lower[ok] + mu[ok]

    return samples.T



from .Gallery import *
from ..plottables import *


class GridGallery(Gallery):
    def __init__(
        self,
        rows=[Declination, Radius],
        cols=[RightAscension, Flux],
        mapsize=(5, 3),
        **kw
    ):
        self.mapsize = mapsize
        self.setup_maps(rows, cols, **kw)

    def setup_maps(self, rows=[], cols=[], map_type=BubbleMap, **kw):

        # create the subplots
        nrows, ncols = len(rows), len(cols)
        figsize = (self.mapsize[0] * len(cols), self.mapsize[1] * len(rows))
        self.fi, self.ax = self.create_subplots(
            nrows=nrows, ncols=ncols, figsize=figsize, **kw
        )

        self.ax = np.atleast_2d(self.ax).reshape((nrows, ncols))

        # attach maps to each ax
        self.maps = {}
        for i, r in enumerate(rows):
            for j, c in enumerate(cols):
                m = map_type(xaxis=c, yaxis=r)
                m.ax = self.ax[i, j]
                self.maps[m.label] = m

    def refine_maps(self):
        """
        Make small changes to Maps, after data are plotted.
        """
        for a in np.atleast_1d(self.ax[:, 1:]).flatten():
            a.set_ylabel("")
        for a in np.atleast_1d(self.ax[:-1, :]).flatten():
            a.set_xlabel("")



from ..maps import *


def clean_maps(maps):
    """
    Make sure that a collection of Maps
    is stored as a dictionary.

    Users might sometimes provide a list
    or sometimes a dictionary. This helper
    makes sure it's a dictionary.
    """
    if isinstance(maps, dict):
        # if it's a dictionary, make su
        maps_as_dictionary = maps
    else:
        # otherwise, assume it's a list
        maps_as_dictionary = {p.label: p for p in maps}

    # make deepy copies of all maps to avoid connections between galleries
    for k, v in maps_as_dictionary.items():
        assert isinstance(v, Map)
        maps_as_dictionary[k] = v.copy()
    return maps_as_dictionary


class Gallery:
    def __init__(
        self,
        maps=[],
        label=None,
        **kw,
    ):
        """
        Initialize a Gallery of Maps.

        The default will display a set of Maps in one
        row or column. For building more complicated
        galleries of maps you might want to start
        from here and overwrite two methods:
        - `setup_maps()` to connect maps to axes
        - `refine_maps()` to fuss with details
        The docstrings for these two methods briefly
        sketch what each is for.

        Parameters
        ----------
        maps : dict, list
            A dictionary or list of Maps that should be
            included in this Gallery. If Maps are not
            specified here, they must be defined inside of
            `.setup_maps()`. If it's a
        label : str, None
            A string to label this Gallery, which will
            mostly appear in filenames for saved plots.
        **kw : dict
            All additional keywords will be passed to
            `.setup_maps()` to decide how the subplots
            will be arranged.
        """

        # convert the maps into a dictionary
        self.maps = clean_maps(maps)

        # create the figure + axes (but leave them empty)
        self.setup_maps(**kw)

        # make sure this gallery has a label
        self.label = label or "+".join([p.label for p in self.maps.values()])

    def create_subplots(
        self,
        nrows=1,
        ncols=1,
        sharex="col",
        sharey="row",
        figsize=(9, 6),
        width_ratios=None,
        height_ratios=None,
        wspace=None,
        hspace=None,
        **kw,
    ):
        """
        This helper just sets some reasonable defaults
        for constructing the figure and axes into
        which Maps will be organized.

        Parameters
        ----------
        nrows : int
            How many rows?
        ncols : int
            How many columns?
        sharex :
            How should the x-axis be shared across maps?
            (default 'col') shares along columns
        sharey : bool, str
            How should the y-axis be shared across maps?
                (default 'row') shares along rows
        figsize : tuple
            How big should the figure be?
        width_ratios : tuple
            How wide should the columns be relative to each other?
        height_ratios : tuple
            How tall should the rows be relative to each other?
        wspace : float
            How much space between columns?
        hspace : float
            How much space between rows?
        """
        return plt.subplots(
            nrows=nrows,
            ncols=ncols,
            sharex=sharex,
            sharey=sharey,
            figsize=figsize,
            constrained_layout=True,
            gridspec_kw=dict(
                width_ratios=width_ratios,
                height_ratios=height_ratios,
                wspace=wspace,
                hspace=hspace,
            ),
        )

    def remove_unused_axes(self):
        """
        If any axes aren't connected to maps, remove them.
        """
        for a in np.atleast_1d(self.ax.flatten()):
            is_used = np.any([a == p.ax for p in self.maps.values()])
            if is_used == False:
                a.axis("off")

    def setup_maps(self, horizontal=True, **kw):
        """
        Create the figure + axes, and
        link the axes to Map objects.

        You might want to write over this for a
        more complicated Gallery of Maps!

        This should create two internal variables:
            `self.fi` = the figure
            `self.ax` = a 1D or 2D grid of axes
        It should also make sure that every map
        to plot in `self.maps` is connected to an
        ax in `self.ax`, either by creating the map
        with a `ax=...` keyword, or by setting the
        `.ax` attribute of an already existing Map.

        Parameters
        ----------
        horizontal : bool
            If True, make a row of Maps.
            If False, make a column of Maps.
        **kw : dict
            All other keywords will be passed to
            `.create_subplots()` for setting up the
            figure and axes. See its docstring.
        """

        # decide rows and columns
        N = len(self.maps)
        self.horizontal = horizontal
        if horizontal:
            rows, cols = 1, N
        else:
            rows, cols = N, 1

        # create the subplots
        self.fi, self.ax = self.create_subplots(nrows=rows, ncols=cols, **kw)

        # attach maps to each ax
        for i, k in enumerate(self.maps):
            self.maps[k].ax = np.atleast_1d(self.ax.flatten())[i]

    def refine_maps(self):
        """
        Make small changes to Maps, after data are plotted.

        You might want to write over this for a
        more complicated Gallery of Maps!

        We might want to plot some extra curves on
        a map, or fuss with its axis labels, or
        nudge its ticks, or add a legend. This
        function gets called after the data have
        been built up into the plot, so those
        modifications can be made.

        By default, this simply removes axis labels
        from
        """
        if self.horizontal:
            for a in self.ax[1:]:
                # plt.setp(a.get_yticklabels(), visible=False)
                a.set_ylabel("")
        else:
            for a in self.ax[:-1]:
                # plt.setp(a.get_xticklabels(), visible=False)
                a.set_xlabel("")

    def build(self, pops, save=False, steps=True, format="png", savefig_kw={}):
        """
        Populate all Maps in a Gallery,
        using a set of populations.

        Parameters
        ----------
        pops : dict
            The populations to plot.
        save : bool
            Should we save the figure(s) to files?
        format : str
            What kind of graphics file should be saved?
            Any format allowed by `plt.savefig` is fine;
            we recommend `png` or `pdf`.
        savefig_kw : dict
            Keywords to pass to `plt.savefig` for saving figures.

        Returns
        -------
        self : Gallery
            In case you want to modify the plot after it's
            generated, this returns the entire Gallery.
        """

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # put data into the axes
            for k, p in self.maps.items():
                p.build(pops=pops, legend=False)

            # remove any empty axes
            self.remove_unused_axes()

            # make small tweaks to the maps
            self.refine_maps()

            # save figure
            if save:
                filename = f"{self.label}"
                plt.savefig(f"{filename}.{format}", **savefig_kw)

        return


class FourPanelTransitGallery(Gallery):
    def __init__(
        self,
        maps=[
            Mass_x_Radius(),
            Flux_x_Radius(),
            StellarRadius_x_PlanetRadius(),
            Distance_x_Radius(),
        ],
        **kw,
    ):
        Gallery.__init__(self, maps=maps, **kw)



from ..imports import *

__all__ = ["plot_zeng", "plot_three_zeng"]


# plot the mass radius models
def plot_zeng(which="75pSi_25pFe", **kw):
    """
    Plot a single mass-radius model from Zeng & Sasselov (2013).

    Parameters
    ----------
    which : str
        A string to indicate which model to plot.

    """

    file = os.path.join(code_directory, f"models/data/zengandsasselov/MR_{which}.csv")

    # load the values
    m, r = np.transpose(np.loadtxt(file, delimiter=","))

    # plot the values
    return plt.plot(m, r, **kw)


def plot_three_zeng(**kw):
    """
    Plot both the rock and the cold H/He models from Seager et al. (2007)
    """
    zkw = dict(linewidth=4, alpha=0.25, zorder=1.0)
    zkw.update(**kw)
    plot_zeng("75pSi_25pFe", label="75% Si, 25% Fe", color="sienna", **zkw)
    plot_zeng("50pSi_50pFe", label="50% Si, 25% 50", color="brown", **zkw)
    plot_zeng("25pSi_75pFe", label="25% Si, 75% Fe", color="maroon", **zkw)



"""
Define a subset of all exoplanets for only those known to transit. 
"""

from ...imports import *
from .exoplanets import *

__all__ = ["TransitingExoplanets"]


class TransitingExoplanets(Exoplanets):
    """
    Transiting Exoplanets from the NASA Exoplanet Archive.
    """

    def __init__(self, **kw):
        Exoplanets.__init__(self, **kw)

        # set the label
        if "label" not in kw:
            self.label = "Transiting Exoplanets"

        # tidy up this population
        self._remove_nontransiting()

    def _remove_nontransiting(self):
        """
        Remove non-transiting planets from the population.

        Returns
        -------
        (modifies the input population in place, but returns nothing)
        """
        ok = self.detected_in_transit() == 1

        self.standard = self.standard[ok]
